# Platformy Technologiczne Powtórzenie

## Platforma Java

Michał Wójcik

---

### Java

**Java** - rodzaj kawy pochodzący z wyspy Jawa znajdującej się w Indonezji.

---

### Platforma Java - elementy

**Java Platform** - platforma softwarowa działająca nad platformami hardwarowymi. Składa się z
dwóch istotnych elementów:

* **JVM** - **Java Virtual Machine**, maszyna wirtualna, środowisko zdolne do wykonywania kodu
  bajtowego uzyskanego w wyniku kompilacji kodu źródłowego w języku Java (i nie tylko),
* **API** - **Application Programming Interface**, duży zbiór gotowych komponentów oprogramowania,
  które zapewniają wiele przydatnych funkcjonalności (np.: przetwarzanie plików, obsługa wątków,
  obsługa połączeń sieciowych itd.).

---

### Platforma Java - standardy

**Java Community Process** - (**JCP**) mechanizm służący do opracowywania standardowych
specyfikacji technicznych dla technologii Java. Każdy może zarejestrować się oraz brać udział
w przeglądaniu i przekazywaniu opinii na temat specyfikacji związanych z platformą Java
(**Java Specification Request** - **JSR**).

W ramach JCP są rozwijane m.in. dwie istotne specyfikacje (JSR):

* **Java Language Specification** - (**JLS**) specyfikacja języka Java.
* **Java Virtual Machine Specification** - (**JVMS**) specyfikacja maszyny wirtualnej Java.

---

### Platforma Java - narzędzia

**Java Language** - obiektowy język programowania stworzony z myślą o platformie Java zgodny
ze specyfikacją języka Java (**JLS**).

**Java Development Kit** - (**JDK**) narzędzia programistyczne dla platformy Java. Implementuje
specyfikację języka Java (**JLS**), specyfikację maszyny wirtualnej Java (**JVMS**) oraz dostarcza
gotowy zbiór komponentów oprogramowania (**Java SE API**).

**Java Runtime Environment** - (**JRE**) środowisko uruchomieniowe dla programów przeznaczonych
na platformę Java. Tworzy instancje maszyny wirtualnej Java (**JVM**) na potrzeby uruchomienia
aplikacji.

---

### Przenośność

**Przenośność** może być rozumiana na różne sposoby.

Przenośność aplikacji **C++** - raz napisana aplikacja może zostać skompilowana pod konkretną
platformę (sprzęt, system operacyjny), gdzie następnie może być uruchamiana.

Przenośność aplikacji **JavaScript** - raz napisana aplikacja może być wykonywana
(interpretowana) na dowolnej platformie, o ile jest tam dostępny interpreter (zazwyczaj w ramach
przeglądarki internetowej).

Przenośność aplikacji **Java** - raz napisane i skompilowane aplikacja może zostać uruchomiona
na dowolnej platformie, o ile jest tam dostępna maszyna wirtualna w odpowiedniej wersji.

---

### Formaty plików

**Formaty plików**:

* pliki `*.java` - pliki zawierające kod źródłowy w języku Java, kompilowane
  do kodu bajtowego maszyny wirtualnej,
* pliki `*.class` - pliki zawierające kod bajtowy maszyny wirtualnej (Java
  bytecode), lista instrukcja zgodna ze zbiorem instrukcji maszyny wirtualnej Java,
* pliki `*.jar` - pliki archiwa zawierające kilka plików `*.class`, metadane (np. plik
  `MANIFEST.MF`) oraz statyczne zasoby (obrazki, dźwięki, konfiguracja itd.).

---

### Kompilator Just-In-Time

**JIT** - **Just-In-Time**:

* kompilator w ramach maszyny wirtualnej Java,
* interpretowanie kodu jest o wiele wolniejsze od wykonania na procesorze,
* aplikacja w pierwszej fazie nie jest kompilowana do kodu natywnego a jedynie interpretowana,
* kompilacja do kodu natywnego jest kosztowna i nieopłacalna dla rzadko używanych metod.

Dwa podstawowe typu kompilatorów JIT:

* **C1** - kompilator kliencki - stara się zoptymalizować i skompilować kod tak szybko jak to
  możliwe,
* **C2** - kompilator serwerowy - analizuje kod przez dłuższy czas w celu wykonania bardziej
  zaawansowanych i kosztownych optymalizacji.

**Kompilacja etapowa** - **Tiered Compilation**:

* kod jest interpretowany i profilowany,
* kod jest kompilowany za pomocą kompilatora C1 i profilowany,
* kod jest kompilowany za pomocą kompilatora C2.

---

### Model pamięci

**Pamięć** w ramach JVM:

* **stos** - **stack** - statyczna alokacja pamięci:
  * zawiera wartości typów prymitywnych dla konkretnej metody oraz referencje do obiektów
    znajdujących się na stercie,
  * dostęp do pamięci w postaci kolejki LIFO,
  * każde wywołanie metody tworzy nową ramkę na stosie,
  * zakończenie metody usuwa ramkę ze stosu;
* **sterta** - **heap** - dynamiczna alokacja pamięci:
  * zawiera obiekty tworzone podczas wykonania aplikacji,
  * młoda generacja - miejsce na wszystkie nowe obiekty, czyszczona przez GC,
  * stara generacja - miejsce na długotrwałe obiekty przeniesione z młodej generacji,
    czyszczona przez GC,
  * stała generacja - metadane maszyny wirtualnej.

**GC** - **Garbage Collector**:

* automatyczne zwalnianie pamięci,
* zliczanie referencji,
* różne algorytmy.

---

### Technologie Java

**Technologie Java**:

* **Java Card**:
  * aktualna wersja 3.2 (2023),
  * przeznaczona na karty elektroniczne (smart card),
  * np.: karty hotelowe;
* **Java Micro Edition (Java ME)**:
  * aktualna wersja 8.3.1 (2018),
  * przeznaczona na urządzenia mobilne i wbudowane,
  * np.: odtwarzacze Blu-ray, czytniki;
* **Java Standard Edition (Java SE)**:
  * aktualna wersja LTS 17 (często używane wersje 8 i 10),
  * tworzenie aplikacji desktopowych;
* **JavaFX**:
  * aktualna wersja 19,
  * do wersji 10 część Java SE, od wersji 11 samodzielny moduł,
  * tworzenie aplikacji okienkowych;

---

### Technologie Java - serwerowe

**Technologie Java**:

* **Java Enterprise Edition (Java EE)**:
  * aktualna (ostatnia) wersja 8 (2017),
  * tworzenie aplikacji serwerowych (webowych i klasy enterprise);
* **Jakarta Enterprise Edition (Jakarta EE)**:
  * następca Java EE,
  * aktualna wersja 10 (2022);

---

### Technologie Java - inne

* **Android**:
  * przeznaczona na urządzenia mobilne i wbudowane,
  * Dalvik, ART.

---

### Różnorodność - IDE i JVM

**Zintegrowane środowiska programistyczne (IDE)**:

* IntelliJ IDEA (subiektywnie zalecane),
* NetBeans,
* Eclipse,
* Android Studio,
* ...

**Maszyny wirtualne**:

* HotSpot (Oracle),
* Eclipse OpenJ9 (IBM),
* GraalVM (Oracle),
* Jikes RVM (Research Virtual Machine),
* ...

---

### Różnorodność - SDK

**Narzędzia dla programistów (SDK)**:

* OpenJDK builds by Oracle (niezalecane),
* Oracle Java SE Development Kit (niezalecane),
* Adoptium Eclipse Temurin (zalecane),
* AdoptOpenJDK (już nieaktywne),
* Azul Zulu (zalecane),
* BellSoft Liberica JDK (zalecane),
* IBM Semeru Runtime (zalecane jeśli OpenJ9),
* Amazon Corretto (zalecane)
* ...

---

### Różnorodność - języki

**Języki**:

* Java,
* Scala,
* Groovy,
* Kotlin,
* Clojure (dialekt Lisp),
* Fantom,
* JRuby (implementacja Ruby),
* Jythoon (implementacja Python),
* Rhino (silnik JavaScript),
* Nashorn (silnik JavaScript),
* ...

---

### Różnorodność - zależności

**Biblioteki**:

* Apache Commons,
* JUnit,
* Mockito,
* ...

**Frameworki**:

* Spring Framework,
* JavaServer Faces,
* Play Framework,
* ...

---

### Pomocne źródła

Książki:

* Cay S. Horstmann, *Java. Podstawy. Wydanie XI*, Helion, 2019.
* Cay S. Horstmann, *Java. Techniki zaawansowane. Wydanie XI*, Helion, 2020.
* Herbert Schildt, *Java. Przewodnik dla początkujących. Wydanie VIII*, Helion, 2020.
* Herbert Schildt, *Java. Kompendium programisty. Wydanie XI*, Helion, 2020.

Dokumentacja:

* Oracle Corporation, *The Java Tutorials*, <https://docs.oracle.com/javase/tutorial/>.
* Oracle Corporation, *JDK 8 Documentation*, <https://docs.oracle.com/javase/8/>.
* Oracle Corporation, *JDK 11 Documentation*, <https://docs.oracle.com/javase/11/>.
* Oracle Corporation, *JDK 17 Documentation*, <https://docs.oracle.com/javase/17/>.

Kursy:

* Baeldung, *Java “Back to Basics” Tutorial*, <https://www.baeldung.com/java-tutorial>.
* Baeldung, *Java*, <https://www.baeldung.com/category/java>.
* w3schools, *Java Tutorial*, <https://www.w3schools.com/java/>.

---

### Pomocne źródła - uzupełnienie

Uzupełnienie informacji:

* <https://www.baeldung.com/jvm-vs-jre-vs-jdk>
* <https://www.baeldung.com/java-jar-manifest>
* <https://www.baeldung.com/graal-java-jit-compiler>
* <https://www.baeldung.com/jvm-tiered-compilation>
* <https://www.baeldung.com/java-stack-heap>
* <https://www.baeldung.com/jvm-garbage-collectors>
