# Platformy Technologiczne

## Budowanie aplikacji

Michał Wójcik

---

### Automatyzacja budowania aplikacji

**Automatyzacja budowania oprogramowania** - automatyzacja procesów związanych z budowaniem
oprogramowania, m.in.:

* kompilacja,
* testowanie,
* generowanie dokumentacji,
* pakowanie,
* ...

**Narzędzia** automatyzujące budowanie:

* Make,
* Rake,
* Cake,
* MSBuild,
* Ant,
* Maven,
* Gradle.

---

### Maven

**Maven** - narzędzie automatyzujące budowę oprogramowania:

* narzędzie cieszące się dużą popularnością w przypadku projektów opartych na platformie Java,
* zorientowane na wtyczki realizujące różne funkcjonalności,
* może odpowiadać za budowanie, testowanie,
* dokumentowanie i wdrażanie projektu,
* konfiguracja w postaci pliku XML,
* nie wymaga żadnego IDE do działania,
* zintegrowany we wszystkich popularnych IDE,
* samodzielne narzędzie możliwe do użycia:
  * z poziomu powłoki,
  * w skryptach,
  * z poziomu automatycznych narzędzi typu (np.: serwer continuous integration).

---

### Struktura projektu

Narzędzie Maven zakłada konkretną strukturę projektu:

* `pom.xml` - plik konfiguracyjny,
* `src`- źródła projektu,
* `target` - wynik budowania projektu,
* `src/main` - główne źródła projektu,
* `src/test` - źródła testowe,
* `java` - kompilowane pliki `.java`,
* `resources` - niekompilowane zasoby, np.: style, pliki graficzne i dźwiękowe itd.

```tex
|-project-name
|--pom.xml
|--src
|---main
|----java
|----resources
|---test
|----java
|----resources
|-target
```

---

### Konfiguracja projektu

```xml
<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 
                             http://maven.apache.org/xsd/maven-4.0.0.xsd">

    <modelVersion>4.0.0</modelVersion>

    <groupId>pl.edu.pg.kask.pt</groupId>
    <artifactId>example</artifactId>
    <version>1.0.0-SNAPSHOT</version>

</project>
```

Każdy projekt (lub biblioteka) jest opisany przez trzy własności:

* `groupId` - identyfikator grupy projektu, zwykle zawiera identyfikator firmy lub grupy stojącej za
  projektem,
* `artifactId` - identyfikator projektu,
* `version` - wersja.

**Artefakt** - produkt budowania projektu opartego na narzędziu Apache Maven.

---

### Wersja języka

Ustawienie wersji dla źródeł oraz wyniku kompilacji:

```xml
<project>
    <properties>
        <maven.compiler.source>17</maven.compiler.source>
        <maven.compiler.target>17</maven.compiler.target>
    </properties>
</project>
```

```xml
<project>
    <properties>
        <maven.compiler.source>11</maven.compiler.source>
        <maven.compiler.target>11</maven.compiler.target>
    </properties>
</project>
```

```xml
<project>
    <properties>
        <maven.compiler.source>1.8</maven.compiler.source>
        <maven.compiler.target>1.8</maven.compiler.target>
    </properties>
</project>
```

---

### Cykle życia

Budowanie oprogramowania z wykorzystaniem narzędzia maven oparte jest na **cyklach** (lifecycles).

Wbudowane cykle życia:

* **default** - umożliwiający budowanie projektu, jego wdrażanie i publikację artefaktów,
* **clean** - umożliwia wyczyszczenie projektu (usunięcie wszystkich plików powstających w cyklu
  default),
* **site** - umożliwiający wygenerowanie statycznych stron HTML z dokumentacją projektu.

Aby wywołać konkretny cykl:

```bash
mvn clean
```

---

### Fazy cyklu

Każdy cykl życia składa się z następujących po sobie faz (ang. phases).

Główne fazy cyklu **default**:

1. **validate** - sprawdzanie poprawności projektu,
2. **compile** - kompilacja kodu źródłowego,
3. **test** - wykonanie testów jednostkowych,
4. **package** - pakowanie skompilowanych klas i zasobów do archiwum (np.: jar),
5. **verify** - testy integracyjne,
6. **install** - instalacja w lokalnym repozytorium:
   * inne lokalne projekty mogą korzystać z zależności;
7. **deploy** - wgranie do zdalnego repozytorium:
   * inni deweloperzy mogą korzystać z zależności.

Aby zbudować aplikację do (włącznie) określonej fazy:

```bash
mvn package
```

---

### Cele faz

Każda faza składa się z listy **celów** (goals) pochodzących od wtyczek.

Wybrane przykłady celów:

| cel              | wtyczka  | faza    |
|------------------|----------|---------|
| compiler:compile | compiler | compile |
| surefile:test    | surefile | test    |
| install:install  | install  | install |
| jar:jar          | jar      | package |

---

### Stworzenie projektu

```bash
mvn archetype:generate \
    -DgroupId=pl.edu.pg.kask.pt \
    -DartifactId=maven-example \
    -DarchetypeGroupId=org.apache.maven.archetypes \
    -DarchetypeArtifactId=maven-archetype-quickstart \
    -DarchetypeVersion=1.4 \
    -DinteractiveMode=false
```

```bash
cd maven-example
mvn package
java -classpath target/maven-example-1.0-SNAPSHOT.jar pl.edu.pg.kask.pt.App
```

**Archetyp** - szablon projektu opartego na narzędziu Apache Maven. Podobnie jak wtyczki i artefakty,
każdy archetyp opisany jest przez `groupId`, `artifactId` i `version`.

Przykładowy standardowych archetyp `maven-archetype-quickstart` jest już dawno nieaktualny.
Istnieje wiele innych archetypów.  Ponadto profesjonalne firmy definiują własne.

---

### Zależności

**Zależności** projektu:

* biblioteki używane w projekcie,
* biblioteki mają też własne zależności, które też mają zależności - drzewo zależności,
* konkretna wersja biblioteki zależy od konkretnych wersji swoich zależności,
* aktualizacja biblioteki wymaga aktualizacji całej gałęzi drzewa zależności.

---

### Zależności w `pom.xml`

**Zależności** - zewnętrzne biblioteki wymagane przez aplikację:

* deklarowane w `pom.xml`,
* ściągane automatycznie z maven central repo,
* możliwość ustawienia prywatnych repozytoriów.

**Scope** - zasięg zależności:

* `compile` - wymagane w czasie kompilacji oraz w czasie działania aplikacji,
* `provided` - jak wyżej, ale dostarczane przez środowisko wykonawcze, np. serwer aplikacji,
* `runtime` - wymagane w czasie działania aplikacji (ale nie w czasie kompilacji)`
* `test` - wymagane w czasie uruchamiania testów jednostkowych,
* `import` - import zależności z innego pliku `pom.xml`,
* `system` - wskazanie lokalizacji w systemie, niezalecane.

---

### Zależności - przykład

```xml
<project>
    <dependencies>
        <dependency>
            <groupId>junit</groupId>
            <artifactId>junit</artifactId>
            <version>4.12</version>
            <scope>test</scope>
        </dependency>

        <dependency>
            <groupId>org.apache.commons</groupId>
            <artifactId>commons-lang3</artifactId>
            <version>3.9</version>
        </dependency>
    </dependencies>
</project>
```

---

### Repozytoria

Wszystkie zależności, wtyczki, archetypy pobierane są z **repozytorium**:

* domyślnie wykorzystywane jest centralne repozytorium: <https://repo1.maven.org/maven2/>,
* subiektywnie polecany graficzny interfejs do wyszukiwania artefaktów: <https://mvnrepository.com/>,
* możliwość definiowania dodatkowych repozytoriów w `pom.xml`,
* możliwość konfigurowania własnych repozytoriów, np.: Nexus Repository OSS.

```xml
<project>
    <repositories>
        <repository>
            <id>repository.jboss.org</id>
            <name>JBoss - Releases Repository</name>
            <url>https://repository.jboss.org/nexus/content/repositories/releases</url>
        </repository>
    </repositories>
</project>
```

---

### Wtyczki - domyslne

Proces budowania realizowany jest za pomocą **wtyczek** realizujących odpowiednie funkcjonalności.
Domyślnie maven dodaje m.in. następujące pluginy:

* `maven-install-plugin` - wgranie artefaktu do lokalnego repozytorium,
* `maven-compiler-plugin` - proces kompilacji,
* `maven-surefire-plugin` - proces testowania,
* `maven-jar-plugin` - tworzenie archiwum jar,
* `maven-deploy-plugin` - wgranie artefaktu do zdalnego repozytorium.

---

### Wtyczki - inne przydatne

Inne przydatne Wtyczki:

* `maven-javadoc-plugin` - wygenerowanie paczki z dokumentacją aplikacji,
* `maven-source-plugin` - wygenerowanie paczki ze źródłami aplikacji,
* `maven-dependency-plugin` - skopiowanie wszystkich zależności aplikacji do katalogu target.

---

### Ustawianie głównej klasy

```xml
<project>
    <build>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-jar-plugin</artifactId>
                <version>3.1.2</version>
                <configuration>
                    <archive>
                        <manifest>
                            <addClasspath>true</addClasspath>
                            <mainClass>pl.edu.pg.kask.pt</mainClass>
                        </manifest>
                    </archive>
                </configuration>
            </plugin>
        </plugins>
    </build>
</project>
```

---

### Generowanie i pakowanie dokumentacji

```xml
<project>
    <build>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-javadoc-plugin</artifactId>
                <version>3.1.1</version>
                <executions>
                    <execution>
                        <id>attach-javadocs</id>
                        <goals>
                            <goal>jar</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>
</project>
```

---

### Pakowanie źródeł

```xml
<project>
    <build>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-source-plugin</artifactId>
                <version>3.1.0</version>
                <executions>
                    <execution>
                        <id>attach-sources</id>
                        <goals>
                            <goal>jar</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>
</project>
```

---

### Generowanie dokumentacji

```xml
<project>
    <reporting>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-javadoc-plugin</artifactId>
                <version>3.1.1</version>
                <configuration>
                    <show>private</show>
                    <nohelp>true</nohelp>
                </configuration>
            </plugin>
        </plugins>
    </reporting>
</project>
```

---

### Profile

**Profile** pozwalają utworzenie niezależnych zestawów wtyczek. Przykładowo:

* *default* - jedynie kompilacja i testowanie,
* *release* - kompilacja, przygotowanie archiwum jar, wygenerowanie dokumentacji oraz archiwum ze
  wszystkimi elementami aplikacji.

```xml
<project>
    <profiles>
        <profil >
            <id>release</id>
            <build>
            </build>
        </profile>
    </profiles>
</project >
```

Aby wykorzystać profil:

```bash
mvn -Prelease package
```

---

### Pomocne źródła

Oficjalna dokumentacja:

* Apache, *Maven in 5 minutes*,
  <https://maven.apache.org/guides/getting-started/maven-in-five-minutes.html>.
* Apache, *Maven documentation*, <https://maven.apache.org/guides/getting-started/index.html>.

Uzupełnienie informacji:

* <https://www.baeldung.com/maven-artifact>
* <https://www.baeldung.com/maven-goals-phases>
* <https://www.baeldung.com/maven-dependency-scopes>
