# Platformy Technologiczne

## Kolekcje i porównywanie obiektów

Michał Wójcik

---

### Kolekcje

**Kolekcja** - dynamiczny (w przeciwieństwie do tablic) kontener, obiekt grupujący elementy.

**Collection framework** - narzędzia służące reprezentacji i manipulacji kolekcjami:

* interfejsy - abstrakcyjne typy reprezentujące kolekcje,
* implementacje - konkretne implementacje interfejsów,
* algorytmy - metody udostępniające użyteczne funkcjonalności (np.: sortowanie).

---

### Rodzaje kolekcji - diagram

![Rodzaje kolekcji](./assets/img/kolekcje.svg)

---

### Rodzaje kolekcji

**Rodzaje kolekcji** (interfejsy):

* `Iterable` - deklaruje możliwość iterowania po elementach, pobranie iteratora lub wykorzystanie w
pętlach `for-each`,
* `Collection` - najbardziej ogólny rodzaj kolekcji, brak ograniczeń typu powtarzalność elementów
lub kolejność,
* `Set` - nie zawiera powtarzających się elementów (takich samych), brak zapewnienia kolejności
  występowania elementów,
* `SortedSet` - automatyczne sortowanie wstawianych elementów,
* `NavigableSet` - możliwość pobierania elementów *większych*/*mniejszych* niż wskazane,
* `List` - sekwencja elementów, może zawierać takie same elementy,
* `Queue` - kolejka FIFO,
* `Deque` - kolejka FIFO, LIFO, lista dwukierunkowa.

---

### Rodzaje map - diagram

![Rodzaje map](./assets/img/mapy.svg)

---

### Rodzaje map

**Rodzaje kolekcji** (interfejsy):

* `Map` - ograniczenia na klucze takie jak na elementy w `Set`, nie zawiera powtarzających się kluczy (takich samych),
  brak zapewnienia kolejności występowania kluczy,
* `SortedMap` - ograniczenia na klucze takie jak na elementy w `SortedSet`, automatyczne sortowanie
  wstawianych kluczy,
* `NavigableMap` - ograniczenia na klucze takie jak na elementy w `NavigableSet`, możliwość
pobierania kluczy "większych"/"mniejszych" niż wskazane,

---

### Operacje na kolekcjach

**Operacje** na pojedynczych elementach w `Collection`:

* `boolean add(Object o)` - oddanie pojedynczego elementu,
* `boolean remove(Object o)` - usunięcie pojedynczego elementu,
* `boolean contains(Object o)` - sprawdzenie, czy element znajduje się w kolekcji.

```jshelllanguage
Collection list = new ArrayList();

list.add("woof");
System.out.println(list.contains("woof")); // true

list.remove("woof");
System.out.println(list.contains("woof")); // false
```

---

### Operacje na mapach

**Operacje** na pojedynczych elementach w `Map`:

* `boolean put(Object key, Object value)` - dodanie pojedynczego elementu,
* `Object get(Object key)` - pobranie wartości,
* `Object remove(Object key)` - usunięcie pojedynczego elementu,
* `boolean containsKey(Object o)` - sprawdzenie, czy klucz znajduje się w kolekcji,
* `boolean containsValue(Object o)` - sprawdzenie, czy wartość znajduje się w kolekcji.

```jshelllanguageJava
Map map = new HashMap();

map.put("wolf", 2);

System.out.println(map.containsKey("wolf")); // true
System.out.println(map.containsValue(2)); // true
```

---

### Przeglądanie kolekcji

Możliwość przeglądania kolekcji za pomocą pętli `for-each`. Podczas przeglądania kolekcji nie można
jej modyfikować (dodawanie/usuwanie elementów, sortowanie, itd.).

```jshelllanguage
Collection list = new ArrayList();

list.add("woof");

for (Object o : list) {
    System.out.println(o);
}
```

---

### Przeglądanie kolekcji - iterator

Możliwość pobrania iteratora z kolekcji:

* `boolean hasNext()` - sprawdza, czy jest jeszcze coś do przejrzenia,
* `Object next()` - przesuwa iterator i zwraca obiekt,
* `void remove()` - usuwa z kolekcji obiekt, na który wskazuje iterator.

Wykorzystanie iteratora pozwala na usuwanie elementów kolekcji podczas przeglądania.

```jshelllanguage
Collection list = new ArrayList();

list.add("woof");

for (Iterator it = list.iterator(); it.hasNext();) {
    Object o = it.next();
    it.remove();
}
```

---

### Przeglądanie mapy

Brak możliwości przeglądania mapy bezpośrednio za pomocą pętli `for-each`. Możliwość skorzystania z
metod:

* `Set keySet()` - zwraca zbiór kluczy,
* `Collection values()` - zwraca kolekcję wartości,
* `Set entrySet()` - zwraca zbiór wpisów (klucz i wartość).

```jshelllanguage
Map map = new HashMap();

map.put("wolf", 2);

for (Object k : map.keySet()) {
    System.out.println(k + " " + map.get(k));
}

for (Object v : map.values()) {
    System.out.println(v);
}

for (Object e : map.entrySet()) {
    System.out.println(((Map.Entry)e).getKey() + " " + ((Map.Entry)e).getValue());
}
```
  
---

### Operacje zbiorcze na kolekcjach

**Operacje zbiorcze** (bulk operations) - operacje na całej kolekcji:

* `boolean containsAll(Collection collection)`,
* `boolean addAll(Collection collection)`,
* `boolean removeAll(Collection collection)`,
* `boolean retainAll(Collection collection)` - usuwa elementy, których nie ma w kolekcji podanej
  jako parameter,
* `void clear()`,
* `int size()`,
* `boolean isEmpty()`.

**Konwersja na tablice**:

* `Object[] toArray()` - zwraca kolekcję jako tablicę typu `Object`,
* `T[] toArray(new T[0])` - zwraca kolekcję jako tablicę typu `T`.

---

### Operacje zbiorcze na mapach

**Operacje zbiorcze** (bulk operations) - operacje na całej mapie:

* `boolean putAll(Map map)`,
* `void clear()`,
* `int size()`,
* `boolean isEmpty()`.

---

### Kolekcje jako typy generyczne

```jshelllanguage
Collection objects = new ArrayList();

objects.add("woof");

for (Object o : objects) {
    String s = (String) o;
}

objects.add(new Integer(2)); // ok
```

```jshelllanguage
Collection<String> strings = new ArrayList<>();

strings.add("woof");

for (String s : strings) {
}

strings.add(new Integer(2)); // compilation error
```

---

### Mapy jako typy generyczne

```jshelllanguage
Map objects = new HashMap();

objects.put("woof", 2);

for (Object k : objects.keySet()) {
    String s = (String) k;
    Integer v = (Integer) objects.get(k);
}

objects.put(new Integer(2), "bark");
```

```jshelllanguage
Map<String, Integer> strings = new HashMap<>();

strings.put("woof", 2);

for (String k : strings.keySet()) {
    Integer v = strings.get(k);
}

for (Map.Entry<String, Integer> e : strings.entrySet()) {
    String k = e.getKey();
    Integer v = e.getValue();
}

strings.put(new Integer(2), "bark"); // compilation error
```

---

### Implementacje kolekcji - zbiory

**Implementacje kolekcji** (klasy):

* `Set`:
  * `HashSet` - implementacja mapy z haszowaniem, brak sortowania i stałej kolejności elementów,
  * `LinkedHashSet` - implementacja mapy z haszowaniem i listy dwukierunkowej, zapewnia kolejność elementów;
* `NavigableSet`:
  * `TreeSet` - implementacja struktury drzewiastej, zapewnia sortownie.

---

### Implementacje kolekcji - listy i kolejki

**Implementacje kolekcji** (klasy):

* `List`:
  * `ArrayList` - implementacja dynamicznie rozszerzalnej tablicy,
  * `Vector` - implementacja dynamicznie rozszerzalnej tablicy (thread-save),
  * `LinkedList` - implementacja listy dwukierunkowej;
* `Queue`:
  * `PriorityQueue` - implementacja kopca, zapewnia sortowanie;
* `Deque`:
  * `LinkedList` - implementacja listy dwukierunkowej.

---

### Implementacje kolekcji - mapy

**Implementacje kolekcji** (klasy):

* `Map`:
  * `HashMap` - implementacja mapy z haszowaniem, brak sortowania i stałej kolejności kluczy,
  * `LinkedHashMap` - implementacja mapy z haszowaniem i listy dwukierunkowej, zapewnia kolejność kluczy;
* `NavigableMap`:
  * `TreeMap` - implementacja struktury drzewiastej, zapewnia sortownie kluczy.

---

### Porównywanie elementów

*Mechanizmy porównywania elementów (np. sprawdzaniu, czy element znajduje się już w kolekcji).*

| Implementacja   | Metoda                                           |
|-----------------|--------------------------------------------------|
| `ArrayList`     | `Object#equals`                                  |
| `Vector`        | `Object#equals`                                  |
| `LinedList`     | `Object#equals`                                  |
| `HashSet`       | `Object#hashCode` oraz `Object#equals`           |
| `LinkedHashSet` | `Object#hashCode` oraz `Object#equals`           |
| `TreeSet`       | `Comparable#compareTo` albo `Comparator#compare` |
| `PriorityQueue` | `Comparable#compareTo` albo `Comparator#compare` |

---

### Porównywanie kluczy i wartości

*Mechanizmy porównywania kluczy.*

| Implementacja   | Metoda                                          |
|-----------------|-------------------------------------------------|
| `HashMap`       | `Object#hashCode` oraz `Object#equals`          |
| `LinkedHashMap` | `Object#hashCode` oraz `Object#equals`          |
| `TreeMap`       | `Comparable#compareTo` lub `Comparator#compare` |

*Mechanizmy porównywania wartości.*

| Implementacja   | Metoda          |
|-----------------|-----------------|
| `HashMap`       | `Object#equals` |
| `LinkedHashMap` | `Object#equals` |
| `TreeMap`       | `Object#equals` |

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/java-arraylist>
* <https://www.baeldung.com/java-tree-set>
* <https://www.baeldung.com/java-hashset>
* <https://www.baeldung.com/java-hashmap>
* <https://www.baeldung.com/java-treemap>
* <https://www.baeldung.com/java-treemap-vs-hashmap>
* <https://www.baeldung.com/java-collections>
