# Platformy Technologiczne

## Gniazdka sieciowe

Michał Wójcik

---

### Adres hosta

Klasa `InetAddress` pozwala na zapisanie adresu IP (Internet Protocol) identyfikującego urządzenie sieciowe.

```jshelllanguage
try {

    InetAddress address = InetAddress.getLocalHost();

    System.out.println(address.getHostName());
    System.out.println(address.getHostAddress());

} catch (UnknownHostException ex) {
    System.err.println(ex);
}
```

```jshelllanguage
try {

    InetAddress address = InetAddress.getByName("pg.edu.pl");

    System.out.println(address.getHostName());
    System.out.println(address.getHostAddress());

} catch (UnknownHostException ex) {
    System.err.println(ex);
}
```

---

### Adres hosta - wiele wartości

Klasa `InetAddress` pozwala także na pobranie kilku wartości (adresów) dla jednej nazwy (np. IPv4 i IPv6).

```jshelllanguage
try {

    for (InetAddress address: InetAddress.getAllByName("pg.edu.pl")) {
        System.out.println(address.getHostName());
        System.out.println(address.getHostAddress());
        System.out.println();
    }

} catch (UnknownHostException ex) {
    System.err.println(ex);
}
```

---

### Adres zasobu

**URL** - adres definiujący adres zasobu w sieci. Składa się z elementów:

* protokół (np.: http, https, ftp),
* opcjonalna nazwa użytkownika,
* opcjonalne hasło, jeśli wymagane do uwierzytelniania,
* adres hosta,
* opcjonalny port (wartości domyślne zdefiniowane dla protokołów),
* ścieżka do pliku (zasobu),
* opcjonalne parametry żądania,
* identyfikator fragmentu.

```tex
protocol://user:password@host:port/path?query#ref
```

---

### Adres zasobu - przykład

Adres zasobu może być reprezentowany przez klasę `URL`.

```jshelllanguage
try {

    URL url = new URL("http://user:pass@host:8080/path?p1=v1&p2=v2#ref");

    System.out.println(url.getProtocol()); // http
    System.out.println(url.getHost()); // host
    System.out.println(url.getPort()); // 8080
    System.out.println(url.getDefaultPort()); // 80
    System.out.println(url.getFile()); // /path?p1=v1&p2=v2
    System.out.println(url.getPath()); // /path
    System.out.println(url.getQuery()); // p1=v1&p2=v2
    System.out.println(url.getRef()); // ref
    System.out.println(url.getAuthority()); // user:pass@host:8080
    System.out.println(url.getUserInfo()); // /user:pass

} catch (MalformedURLException ex) {
    System.err.println(ex);
}
```

---

### Otwieranie połączenia

Bezpośrednio za pomocą klasy `URL` możliwe jest otwarcie połączenia, a następnie pobranie strumieni bajtowych.

```jshelllanguage
try {

    URL url = new URL("...");

    try {

        URLConnection connection = url.openConnection();

        try (
                InputStream is = connection.getInputStream();
                OutputStream os = connection.getOutputStream()
        ) {

            //...

        }

    } catch (IOException ex) {
        System.err.println(ex);
    }

} catch (MalformedURLException ex) {
    System.err.println(ex);
}
```

---

### Gniazdko sieciowe

**Gniazdko sieciowe** - podstawowy mechanizm komunikacji dostępny w aplikacjach. Reprezentuje
dwukierunkowy (nadawanie i odbieranie danych) punkt końcowy połączenia.

Klasa `Socket` - podstawowy element w komunikacji sieciowej dla technologii Java. Reprezentuje punkt
końcowy połączenia.

```jshelllanguage
try {

    InetAddress address = InetAddress.getByName("pg.edu.pl");
    
    try (Socket socket = new Socket(address, 80)) {
        System.out.println("working");
    }
    
} catch (IOException ex) {
    System.err.println(ex);
}
```

---

### Nasłuchiwanie na połączenia

Klasa `ServerSocket` - reprezentuje serwer:

* pozwala na akceptowanie przychodzących połączeń,
* zwraca nowy obiekt `Socket` będący końcem zestawionego połączenia.

```jshelllanguage
try (ServerSocket server = new ServerSocket(8080)) {

    try (Socket socket = server.accept()) {

        //...

    }

} catch (IOException ex) {
    System.err.println(ex);
}
```

---

### Nawiązywanie połączenia

Nawiązywanie połączenia realizuje się bezpośrednio za pomocą obiektu `Socket`.

```jshelllanguage
try (Socket client = new Socket("localhost", 8080)) {

    //...

} catch (IOException ex) {
    System.err.println(ex);
}
```

---

### Połączenie - strumienie

Obiekt `Socket` zwraca podstawowe strumienie bajtowe, na które może być nałożony dowolny dekorator.

```jshelllanguage
try (Socket client = new Socket("localhost", 8080)) {

    try (
            InputStream is = client.getInputStream();
            OutputStream os = client.getOutputStream()
    ) {

        //...
        
    }

} catch (IOException ex) {
    System.err.println(ex);
}
```

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/java-url>
* <https://www.baeldung.com/a-guide-to-java-sockets>
