# Platformy Technologiczne

## Testowanie oprogramowania

Michał Wójcik

---

### Proces testowania oprogramowania

**Testowanie oprogramowania** - jeden z procesów wytwarzania oprogramowania.

Testowanie ma na **celu**:

* *weryfikację oprogramowania* - czy wytwarzane oprogramowanie jest zgodne ze specyfikacją.
* *walidacja oprogramowania* - czy oprogramowanie jest zgodne z oczekiwaniami użytkownika.

Testowanie oprogramowania może być wdrożone w dowolnym momencie wytwarzania oprogramowania (w
zależności od stosowanej metody).

---

### Poziomy testowania

**Poziomy testowania**:

* *testy jednostkowe* - testowanie kodu na poziomie klas,
* *testy integracyjne* - testowanie aplikacji jako całości oraz integracji z innymi elementami
* *testy systemowe* - testy przeprowadzane na kompletnie zintegrowanym systemie,
* *testy akceptacyjne* - w których możemy wyróżnić również specyficzne dla oprogramowania
  komercyjnego testy alfa i testy beta

Istnieją również inne podziały ze względu na umiejscowienie lub przeprowadzanie testów.

---

### Ograniczenia testowanie

**Ograniczenia** testowania:

* programu nie da się przetestować całkowicie,
* testy nie udowodnią braku błędów,
* bardzo często specyfikacja produktu nie jest gotowa co utrudnia testowanie,
* często błędy nie są poprawiane,
* błędy zawsze będą występować w aplikacji.

---

### Testowanie ręczne i automatyczne

**Testowanie ręczne**:

* powolne,
* kosztowne,
* nudne,
* powtarzalność trudna do uzyskania.

**Testowanie automatyczne**:

* szybkie,
* powtarzalne,
* wykonywane podczas budowania aplikacji,
* ułatwia reagowanie na zmiany.

---

### Testy jednostkowe

**Testy jednostkowe (unit tests)**:

* weryfikacja poprawności działania pojedynczych elementów aplikacji,
* wykonywane automatycznie,
* przygotowanie testu polega na napisaniu kodu testującego sprawdzającego kod aplikacji,
* mogą być realizowane podczas procesu budowania aplikacji,
* dokumentują oczekiwane zachowanie testowanego komponentu,
* fakt, że każdy komponent działa poprawnie nie znaczy że całość działa poprawnie.

---

### Przypadki testowe

**Testowane zagadnienia**:

* **happy path** - (poprawne przebiegi) normalna ścieżka przetwarzania,
* **corner case** - (przypadki graniczne) krańce przedziałów dozwolonych wartości:
  * np.: min/max;
* **unhappy path** - przypadki wystąpienia błędów,  
  * np. błędne dane wejściowe.

---

### Biblioteki

Wybrane **biblioteki**:

* wykonywanie testów:
  * JUnit (najpopularniejsza),
  * TestNG;
* asercje:
  * AssertJ (subiektywnie polecana),
  * Hamcrest;
* mokowanie / atrapy:
  * Mockito (najpopularniejsza),
  * PowerMockito,
  * EasyMock.

---

### Przygotowanie testów jednostkowych

**Kod testujący**:

* aplikacje testujemy z wykorzystaniem nowych klas w języku Java, które zawierają kod testujący,
* dane wejściowe przygotowujemy z wykorzystaniem bibliotek tworzących *atrapy* obiektów,
* wyniki działania sprawdzamy z wykorzystaniem *asercji*.

---

### Struktura projektu

```tex
|-project-name
|--pom.xml
|--src
|---main
|----java
|----resources
|---test <--right here
|----java
|----resources
|--target
```

Testy nie wchodzę w część aplikacji dystrybuowanej do klienta. W przypadku projektów opartych na
narzędziu Maven, klasy testowe, jak i zasoby testowe powinny znajdować się w odpowiednim katalogu,
który jest pomijany podczas przygotowania paczki do dystrybucji.

---

### JUnit - zależność

```xml
<dependency>
    <groupId>org.junit.jupiter</groupId>
    <artifactId>junit-jupiter-api</artifactId>
    <version>5.9.2</version>
    <scope>test</scope>
</dependency>
```

Zależności wykorzystywane wyłącznie podczas testów powinny znajdować się w odpowiednim zasięgu.

Uruchomienie testów jest możliwe za pomocą odpowiedniej fazy domyślnego cyklu narzędzia Maven:

```bash
mvn test
```

Lub z poziomu IDE uruchamiając testy jako całość, pojedynczą klasę testową lub pojedynczy test.

---

### Przykładowy test

Klasa testowana:

```java
public class Calculator {

    public int add(int a, int b) {
        return a - b;
    }

}
```

Klasa testująca:

```java
public class CalculatorTest {

    @Test
    public int add() {
        Calculator calc = new Calculator();
        int result = calc.add(2, 2);
        assert result == 4 : "should be 4";
    }

}
```

Wykorzystanie słowa kluczowego `assert` powoduje rzucenie błędu `AssertionError` w sytuacji gdy nie
jest spełniony warunek logiczny. Opcjonalnie można ustawić wiadomość jaka zostanie przypisana do
rzuconego błędu.

---

### Struktura testu - GWT

Kod testowy w metodzie możemy zorganizować z wykorzystaniem zasady ***GWT***:

* **Given** - przygotowanie danych wejściowych oraz atrap,
* **When** - wywołanie testowanego kodu,
* **Then** - zweryfikowanie działania metody oraz jej wyniku.

```jshelllanguage
@Test
public void add() {

  //Given:
  Calculator calc = new Calculator ();

  //When:
  int result = calc.add(2, 2);

  //Then:
  assert result == 4 : "should be 4"; 

}
```

Komentarze w kodzie można pominąć zachowując odpowiednie formatowanie.

---

### Struktura testu - AAA

Kod testowy w metodzie możemy zorganizować z wykorzystaniem zasady ***AAA***:

* **Arrange** - przygotowanie danych wejściowych oraz atrap,
* **Act** - wywołanie testowanego kodu,
* **Assert** - zweryfikowanie działania metody oraz jej wyniku.

```jshelllanguage
@Test
public void add() {

  //Arrange:
  Calculator calc = new Calculator ();

  //Act:
  int result = calc.add(2, 2);

  //Assert:
  assert result == 4 : "should be 4";

}
```

Komentarze w kodzie można pominąć, zachowując odpowiednie formatowanie.

---

### Nazewnictwo testów

**Nazewnictwo testów**:

* klasa testująca zazwyczaj nazywa się tak samo, jak klasa testowana z przyrostkiem `Test`,
* metody testujące nazywają się podobnie do metod testowych, przy czym nazwa zawiera również
  informacje na temat przypadku, jaki jest testowany,
* metoda testująca powinna testować tylko jeden przypadek, każda ścieżka w kodzie powinna być
  testowana przez osobną metodę.

Istnieje wiele konwencji nazewnictwa testów np:

* `MethodName_StateUnderTest_ExpectedBehavior`
  * `isRunning_EngineStarted_True`
* `Should_ExpectedBehavior_When_StateUnderTest`
  * `Should_ReturnTrue_When_EngineIsStarted`

Inne popularne sposoby nazewnictwa: <https://dzone.com/articles/7-popular-unit-test-naming>.

---

### Dobre praktyki

**Dobre praktyki**:

* jedna metoda testowa testuje tylko jeden przypadek,
* testy nie powinny zależeć od siebie nawzajem,
* metody testowe powinny być proste i czytelne:
  * jeśli nie jest to koniecznie, warto unikać konstrukcji warunkowych i pętli;
* testy nie powinny bazować na stanie środowiska, w którym wykonuje się aplikacja.

---

### Test Driven Development (TDD)

***Test driven development*** - zwinna metodyka wytwarzania oprogramowania polegająca na
wielokrotnym powtarzaniu następujących kroków:

* **dodanie testu** - przygotowanie metod testujących na podstawie specyfikacji,
* **uruchomienie wszystkich testów** - nowy test powinien się nie powieść, co oznacza, że testowana
  jest funkcjonalność, której jeszcze nie ma,
* **przygotowanie kodu przechodzącego test** - kod nie musi być dobry, powinien jedynie
  działać poprawnie,
* **uruchomienie wszystkich testów** - jeśli nadal występują błędy, kod musi być poprawiony,
* **refaktoryzacja kodu** - poprawienie kodu,
* **uruchomienie wszystkich testów** - po refaktoryzacji sprawdzamy, czy testy nadal przechodzą.

---

### JUnit - adnotacje

Adnotacje z pakietu `org.junit.jupiter.*`.

| Adnotacja     | Opis                                                                                                                  |
|---------------|-----------------------------------------------------------------------------------------------------------------------|
| `@Test`       | Identyfikuje metodę jako metodę testową.                                                                              |
| `@BeforeEach` | Identyfikuje metodę która będzie wykonywana przed każdym testem. Przygotowanie środowiska np zainicjalizowanie klasy. |
| `@AfterEach`  | Identyfikuje metodę która będzie wykonywana po każdym teście. Posprzątanie po teście np usunięcie plików z dysku.     |
| `@BeforeAll`  | Podobnie jak `@Before` ale metoda wykonywana jest tylko raz dla klasy testowej.                                       |
| `@AfterAll`   | Podobnie jak `@After` ale metoda wykonywana jest tylko raz dla całej klasy testowej.                                  |
| `@Disabled`   | Umieszczony nad klasą testową wyłącza test.                                                                           |

---

### JUnit - mechanizm asercji

JUnit posiada prosty wbudowany mechanizm sprawdzania wyników (asercji)
`org.junit.jupiter.api.Assertions`:

* `assertEquals`, `assertNotEquals` - sprawdzenie, czy obiekty są takie same,
* `assertNull`, `assertNotNull` - sprawdzenie, czy zmienna wskazuje na null,
* `assertTrue`, `assertFalse` - sprawdzanie, czy zmienna ma wartość `true`,
* `assertArrayEquals` - sprawdzenie, czy tablice są takie same,
* `assertThrows` - sprawdzanie, czy fragment kodu rzuci wyjątek.

```java
import static org.junit.jupiter.api.Assertions.assertEquals;
```

```jshelllanguage
String expected = "woof";
String actual = "woof";
assertEquals(expected, actual, "Should be equal");
```

Inne biblioteki pozwalające na sprawne pisanie asercji:

* Hamcrest,
* AssertJ.

---

### Mocki

**Atrapy** (Mocks) - podmienianie nietestowanego obiektu na przygotowaną atrapę:

* służy testowaniu metod korzystających z nietestowanych elementów:
  * strumienie do zewnętrznych zasobów,
  * połączenie z bazą danych,
  * połączenie ze zdalnym serwerem,
* specjalnie zbudowane obiekty realizujące zaimplementowane (przewidywalne) zachowanie,
* popularne biblioteki:
  * `EasyMock`,
  * `Mockito`,
  * `PowerMockito` (`Mockito` na sterydach).

---

### Mockito - zależność

```xml
<dependency>
    <groupId>org.mockito</groupId>
    <artifactId>mockito-junit-jupiter</artifactId>
    <version>5.3.0</version>
    <scope>test</scope>
</dependency>
```

---

### Przekazanie atrapy - kod testowany

Nie testowany interfejs:

```java
public interface DataSource {
    String next();
}
```

Testowana metoda (do działania wymaga instancji interfejsu):

```java
public class DataReader {

    public List<String> readData(DataSource source) {
        List<String> data = new ArrayList<>();
        String line;
        while ((line = source.next()) != null) {
            data.add(line);
        }
        return data;
    }

}
```

---

### Przekazanie atrapy - kod testujący

Testująca metoda:

```jshelllanguage
@Test
public void readData() {
    DataSource source = mock(DataSource.class);
    when(source.next()).thenReturn("woof").thenReturn("woof").thenReturn(null);
    DataReader reader = new DataReader();
    
    List<String> actual = reader.readData(source);

    assertThat(actual).containsExactly("woof", "woof");
}
```

Aby nie testować kilku rzeczy na raz (metody `readData` i implementacji `DataSource`) wykorzystywana
jest atrapa o przewidywalnym zachowaniu.

---

### Tworzenie mocków

```jshelllanguage
import org.mockito.Mockito;
```

```jshelllanguage
DataSource source = Mockito.mock(DataSource.class);
Mockito.when(source.next()).thenReturn("woof").thenReturn("woof").thenReturn(null);
```

```jshelllanguage
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
```

```jshelllanguage
DataSource source = mock(DataSource.class);
when(source.next()).thenReturn("woof").thenReturn("woof").thenReturn(null);
```

---

### Mockito - weryfikacja wywołania metod

```jshelllanguage
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
```

```jshelllanguage
@Test
public void readData() {
    DataSource source = mock(DataSource.class);
    when(source.next()).thenReturn("woof").thenReturn("woof").thenReturn(null);
    DataReader reader = new DataReader();
    
    List<String> actual = reader.readData(source);

    assertThat(actual).containsExactly("woof", "woof");
    verify(source, times(3)).next();
}
```

Oprócz stworzenia atrapy istnieje również możliwość sprawdzenia, czy atrapy zostały wywołane w
poprawny sposób np. Czy parametry, jakie zostały przekazane do atrap, spełniają wymagania.

Do sprawdzenia wykorzystuje się metodę `Mockito.verify`.

---

### Weryfikacja wywołania - kod testowany

Nietestowany interfejs:

```java
public interface DataWriter {
    void write(String value);
}
```

Klasa modelowa:

```java
@AllArgsConstructor
@Getter
public class Elf {
    private String name;
    private int age;
}
```

Testowana klasa:

```java
@AllArgsConstructor
public class ElfWriter {

    private DataWriter writer;

    public void write(Elf elf) {
        writer.write(elf.getName());
        writer.write(Integer.toString(elf.getAge()));
    }
}
```

---

### Weryfikacja wywołania - kod testujący

Metoda testująca:

```jshelllanguage
@Test
public void writeElf() {
    DataWriter writer = mock(DataWriter.class);
    ElfWriter elfWriter = new ElfWriter(writer);
    Elf elf = new Elf("Legolas", 120);

    elfWriter.write(elf);

    verify(writer).write(eq("Legolas"));
    verify(writer).write(eq("120"));
}
```

---

### Pomocne źródła

Oficjalna dokumentacja:

* *JUnit 5 User Guide*, <https://junit.org/junit5/docs/5.9.2/user-guide/>.
* *JUnit 5.9.2 API*, <https://junit.org/junit5/docs/5.9.2/api/>.
* *AssertJ - fluent assertions java library*, <https://assertj.github.io/doc/>.
* *AssertJ Core 3.24.2 API*, <https://javadoc.io/doc/org.assertj/assertj-core/3.24.2/index.html>.
* *Hamcrest Tutorial*, <https://hamcrest.org/JavaHamcrest/tutorial>.
* *Hamcrest 2.2 API*, <https://hamcrest.org/JavaHamcrest/javadoc/2.2/>.
* *Mockito wiki*, <https://github.com/mockito/mockito/wiki/>.
* *Mockito 5.3.0 API*, <https://javadoc.io/doc/org.mockito/mockito-core/5.3.0/org/mockito/Mockito.html>.

---

### Pomocne źródła cd

Uzupełnienie informacji:

* <https://www.baeldung.com/junit-5>
* <https://www.baeldung.com/junit-5-test-annotation>
* <https://www.baeldung.com/junit-before-beforeclass-beforeeach-beforeall>
* <https://www.baeldung.com/junit-assert-exception>
* <https://www.baeldung.com/junit-assertions>
* <https://www.baeldung.com/mockito-junit-5-extension>
* <https://www.baeldung.com/introduction-to-assertj>
* <https://www.baeldung.com/assertJ-java-8-features>
