# Platformy Technologiczne

## Zrównoleglanie operacji

Michał Wójcik

---

### Fork-Join - Java Threads

Model Fork-Join można zaimplementować z wykorzystaniem bazowych mechanizmów związanych z wątkami:

* stworzenie nowego obiektu `Thread` dla pojedynczej części zadania,
* wykorzystanie `Thread#join` do poczekania na wykonanie wątku.

---

### Fork-Join Framework

**Fork-Join Framework** - framework pozwalający na zlecanie wykonania zadań zgodnych z modelem Fork-Join:

* definicja puli wątków,
* definicja zadania.

---

### Pula wątków

Domyślna współdzielona pula wątków:

```jshelllanguage
ForkJoinPool pool = ForkJoinPool.commonPool();
```

Własna pula wątków:

```jshelllanguage
ForkJoinPool pool = new ForkJoinPool(limit);
```

Pule wątków nie zwalniają zasobów automatycznie ani nie są usuwane przez GC. Zamykanie puli wątków:

```jshelllanguage
pool.shutdown();
```

---

### Wykonanie zadania

Wykorzystanie interfejsu `Runnable`:

```jshelllanguage
ForkJoinPool pool = new ForkJoinPool(2);

pool.execute(() -> System.out.println("woof"));
```

---

### Opis zadania

`ForkJoinTask<V>` - bazowa klasa opisująca zadanie wykonywane w ramach puli. Zazwyczaj korzysta się z dwóch podtypów:

* `RecursiveAction` - dla zadań niezwracających wyniku,
* `RecursiveTask<V>` - dla zadań zwracających wynik.

---

### Uruchomienie zadania

```java
class WoofTask extends RecursiveTask<String> {

    @Override
    protected String compute() {
        return "woof";
    }

}
````

```jshelllanguage
ForkJoinPool pool = new ForkJoinPool(1);

WoofTask task = new WoofTask();

pool.execute(task);

System.out.println(task.join());
```

---

### Zadnia rekurencyjne

```java
@AllArgsConstructor
class WoofTask extends RecursiveTask<String> {

    int limit;

    @Override
    protected String compute() {
        if (limit == 0) {
            return "woof";
        } else {
            List<WoofTask> tasks = List.of(
                    new WoofTask(limit - 1),
                    new WoofTask(limit - 1)
            );

            return ForkJoinTask
                    .invokeAll(tasks)
                    .stream()
                    .map(ForkJoinTask::join)
                    .collect(Collectors.joining(" "));
        }
    }

}
```

---

### Równoległe strumienie

Operacje w ramach Stream API można wykonywać równoległe w ramach dostępnej puli wątków.

```jshelllanguage
Collection<String> list = List.of("ookami", "kitsune");

list.stream()
        .parallel()
        .forEach(System.out::println);
```

---

### Własna pula

```jshelllanguage
ForkJoinPool pool = new ForkJoinPool(1);
Collection<String> list = List.of("ookami", "kitsune");

try {

    pool.submit(() -> {
        list.stream()
                .parallel()
                .forEach(System.out::println);
    }).get();

} catch (InterruptedException | ExecutionException e) {
    
}
```

---

### Pomocne źródła

Uzupełnienie informacji:

* <https://www.baeldung.com/thread-pool-java-and-guava>
* <https://www.baeldung.com/java-executor-service-tutorial>
* <https://www.baeldung.com/java-fork-join>
* <https://www.baeldung.com/java-when-to-use-parallel-stream>
